"""SAR Processing for Python

``pysarpro`` (a.k.a. ``pysarpro``) is a collection of algorithms for SAR image
processing.

The main package of ``pysarpro`` only provides a few utilities for converting
between image data types; for most features, you need to import one of the
following subpackages:

Subpackages
-----------
data
    Test images and example data.
io
    Reading, saving, and displaying images and video.
util
    Generic utilities.

Utility Functions
-----------------
img_as_float
    Convert an image to floating point format, with values in [0, 1].
    Is similar to `img_as_float64`, but will not convert lower-precision
    floating point arrays to `float64`.
img_as_float32
    Convert an image to single-precision (32-bit) floating point format,
    with values in [0, 1].
img_as_float64
    Convert an image to double-precision (64-bit) floating point format,
    with values in [0, 1].
img_as_uint
    Convert an image to unsigned integer format, with values in [0, 65535].
img_as_int
    Convert an image to signed integer format, with values in [-32768, 32767].
img_as_ubyte
    Convert an image to unsigned byte format, with values in [0, 255].
img_as_bool
    Convert an image to boolean format, with values either True or False.
dtype_limits
    Return intensity limits, i.e. (min, max) tuple, of the image's dtype.

"""

__version__ = '0.1.3'

from ._shared.version_requirements import ensure_python_version

ensure_python_version((3, 8))

import lazy_loader as lazy

__getattr__, __lazy_dir__, _ = lazy.attach_stub(__name__, __file__)


def __dir__():
    return __lazy_dir__() + ['__version__']


# Logic for checking for improper install and importing while in the source
# tree when package has not been installed inplace.
# Code adapted from scikit-learn's __check_build module.
_INPLACE_MSG = """
It appears that you are importing a local pysarpro source tree. For
this, you need to have an inplace install. Maybe you are in the source
directory and you need to try from another location."""

_STANDARD_MSG = """
Your install of pysarpro appears to be broken.
Try re-installing the package following the instructions at:
https://pol-insar.github.io/docs/stable/user_guide/install.html"""


def _raise_build_error(e):
    # Raise a comprehensible error
    import os.path as osp

    local_dir = osp.split(__file__)[0]
    msg = _STANDARD_MSG
    if local_dir == "pysarpro":
        # Picking up the local install: this will work only if the
        # install is an 'inplace build'
        msg = _INPLACE_MSG
    raise ImportError(
        f"{e}\nIt seems that pysarpro has not been built correctly.\n{msg}"
    )


try:
    # This variable is injected in the __builtins__ by the build
    # process. It used to enable importing subpackages of pysarpro when
    # the binaries are not built
    __pysarpro_SETUP__
except NameError:
    __pysarpro_SETUP__ = False

if __pysarpro_SETUP__:
    import sys

    sys.stderr.write('Partial import of pysarpro during the build process.\n')
    # We are not importing the rest of the scikit during the build
    # process, as it may not be compiled yet
else:
    try:
        from ._shared import geometry

        del geometry
    except ImportError as e:
        _raise_build_error(e)

    # Legacy imports into the root namespace; not advertised in __all__
    from .data import data_dir
    from .util.dtype import (
        dtype_limits,
        img_as_bool,
        img_as_float,
        img_as_float32,
        img_as_float64,
        img_as_int,
        img_as_ubyte,
        img_as_uint,
    )
    from .util.lookfor import lookfor


if 'dev' in __version__:
    # Append last commit date and hash to dev version information, if available

    import os.path
    import subprocess

    try:
        p = subprocess.Popen(
            ['git', 'log', '-1', '--format="%h %aI"'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=os.path.dirname(__file__),
        )
    except FileNotFoundError:
        pass
    else:
        out, err = p.communicate()
        if p.returncode == 0:
            git_hash, git_date = (
                out.decode('utf-8')
                .strip()
                .replace('"', '')
                .split('T')[0]
                .replace('-', '')
                .split()
            )

            __version__ = '+'.join(
                [tag for tag in __version__.split('+') if not tag.startswith('git')]
            )
            __version__ += f'+git{git_date}.{git_hash}'

from pysarpro._shared.tester import PytestTester  # noqa: F401,W203

test = PytestTester(__name__)
del PytestTester



import os

import pytest

import pysarpro.data as data
from pysarpro._shared.testing import assert_equal
from pysarpro.data._fetchers import _image_fetcher


def test_download_all_with_pooch():
    # jni first wrote this test with the intention of
    # fully deleting the files in the data_dir,
    # then ensure that the data gets downloaded accordingly.
    # hmaarrfk raised the concern that this test wouldn't
    # play well with parallel testing since we
    # may be breaking the global state that certain other
    # tests require, especially in parallel testing

    # The second concern is that this test essentially uses
    # a lot of bandwidth, which is not fun for developers on
    # lower speed connections.
    # https://github.com/Pol-InSAR/pysarpro/pull/4666/files/26d5138b25b958da6e97ebf979e9bc36f32c3568#r422604863
    data_dir = data.data_dir
    if _image_fetcher is not None:
        data.download_all(pattern=r'^data/01-sar')
        assert len(os.listdir(data_dir)) >= 1
    else:
        with pytest.raises(ModuleNotFoundError):
            data.download_all()


def test_astronaut():
    """Test that "astronaut" image can be loaded."""
    astronaut = data.astronaut()
    assert_equal(astronaut.shape, (512, 512, 3))


@pytest.mark.parametrize(
    'function_name',
    [
        'file_hash',
    ],
)
def test_fetchers_are_public(function_name):
    # Check that the following functions that are only used indirectly in the
    # above tests are public.
    assert hasattr(data, function_name)



#!/usr/bin/env python

import sys

screen_width = 50

print('*' * screen_width)

if len(sys.argv) > 1:
    header = ' '.join(sys.argv[1:])
    header = header.replace('.', ' ')
    print('*', header.center(screen_width - 4), '*')
    print('*' * screen_width)



#!/usr/bin/env python
# Requires package 'editdistance'

# A mailmap file is used (by GitHub and other tools) to associate multiple
# commit emails with one user.  This helps to count number of commits,
# contributors, etc.

import subprocess
import shlex
import numpy as np
from collections import defaultdict

from editdistance import eval as dist

threshold = 5


def call(cmd):
    return subprocess.check_output(shlex.split(cmd), text=True).split('\n')


def _clean_email(email):
    if '@' not in email:
        return

    name, domain = email.split('@')
    name = name.split('+', 1)[0]

    return f'{name}@{domain}'.lower()


call("rm -f .mailmap")
authors = call("git log --format='%aN::%aE'")

names, emails = [], []

for name, email in (author.split('::') for author in authors if author.strip()):
    if email not in emails:
        names.append(name)
        emails.append(email)

N = len(names)
D = np.zeros((N, N)) + np.infty

for i in range(1, N):
    for j in range(i):
        D[i, j] = dist(names[i], names[j])

for i in range(N):
    (dupes,) = np.where(D[:, i] < threshold)
    for j in dupes:
        names[j] = names[i]

mailmap = defaultdict(set)
for name, email in zip(names, emails):
    email = _clean_email(email)
    if email:
        mailmap[name].add(email)

for key, value in list(mailmap.items()):
    if len(value) < 2 or (len(key.split()) < 2):
        mailmap.pop(key)

entries = []
for name, emails in mailmap.items():
    entries.append([name])
    entries[-1].extend([f'<{email}>' for email in emails])

entries = sorted(entries, key=lambda x: x[0].split()[-1])
for entry in entries:
    print(' '.join(entry))



from tempfile import NamedTemporaryFile

import numpy as np
from pytest import importorskip

from pysarpro._shared.testing import TestCase, assert_array_almost_equal
from pysarpro.io import imread, imsave, reset_plugins, use_plugin

importorskip('imread')


def setup():
    use_plugin('imread')


def teardown():
    reset_plugins()


class TestSave(TestCase):
    def roundtrip(self, x, scaling=1):
        with NamedTemporaryFile(suffix='.png') as f:
            fname = f.name

        imsave(fname, x)
        y = imread(fname)

        assert_array_almost_equal((x * scaling).astype(np.int32), y)

    def test_imsave_roundtrip(self):
        dtype = np.uint8
        np.random.seed(0)
        for shape in [(10, 10), (10, 10, 3), (10, 10, 4)]:
            x = np.ones(shape, dtype=dtype) * np.random.rand(*shape)

            if np.issubdtype(dtype, np.floating):
                yield self.roundtrip, x, 255
            else:
                x = (x * 255).astype(dtype)
                yield self.roundtrip, x



import sys

import numpy as np


def lookfor(what):
    """Do a keyword search on pysarpro docstrings.

    Parameters
    ----------
    what : str
        Words to look for.

    Examples
    --------
    >>> import pysarpro
    >>> pysarpro.lookfor('regular_grid')  # doctest: +SKIP
    Search results for 'regular_grid'
    ---------------------------------
    pysarpro.lookfor
        Do a keyword search on pysarpro docstrings.
    pysarpro.util.regular_grid
        Find `n_points` regularly spaced along `ar_shape`.
    """
    return np.lookfor(what, sys.modules[__name__.split('.')[0]])



"""Provides the possibility to work with RAT files, widely used by DLR-HR institute and some others."""

import copy
import ctypes
import mmap
import os
import warnings
from functools import reduce

import numpy as np

__all__ = [
    'rrat',
    'mrrat',
    'srat',
]

try:
    import lxml.etree as ET
except ImportError:
    warnings.warn(
        'Failed to import ET from "lxml.etree". The Py2XML class will not work.',
        ImportWarning,
    )

try:
    # import importlib_resources
    from importlib import resources

    from mako.template import Template

    # from pkg_resources import resource_string
except ImportError:
    warnings.warn(
        'Failed to import pkg_resources and/or mako. Writing geo-Envi-headers for RAT files will not work.',
        ImportWarning,
    )

red = "\033[91m"
endc = "\033[0m"


def rrat(filename, **kwargs):
    """Read an entire RAT file and return it as a numpy array.

    Parameters
    ----------
    filename : str
        The path to the RAT file to be read.
    **kwargs : keyword arguments, optional
        Additional options for reading the RAT file.

    Returns
    -------
    numpy.ndarray
        The contents of the RAT file as a numpy array.


    """

    rat_file = RatFile(filename)
    return rat_file.read(**kwargs)


def mrrat(filename, **kwargs):
    """Read an entire RAT file and return it as a memory-mapped numpy array.

    Convenient for reading big files, especially for blockwise processing.
    This function works faster than rrat, but the disadvantage is that the
    array is read-only.

    Parameters
    ----------
    filename : str
        The path to the RAT file to be read.
    **kwargs : keyword arguments, optional
        Additional options for reading the RAT file.

    Returns
    -------
    numpy.memmap
        A memory-mapped view of the RAT file contents as a numpy array.

    """

    rat_file = RatFile(filename)
    return rat_file.mread(**kwargs)


def srat(filename, array, **kwargs):
    """Write a numpy ndarray into a RAT file.

    Parameters
    ----------
    filename : str
        The path to the RAT file to be written.
    array : numpy.ndarray
        The numpy array to be written into the RAT file.
    **kwargs : keyword arguments, optional
        Additional options for writing the RAT file.

    Examples
    --------
    >>> data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> srat("output.rat", data)

    """

    rat_file = RatFile(filename)
    rat_file.write(array, **kwargs)


class RatHeaderRat(ctypes.Structure):
    """
    A base class to store RAT header's attributes.

    The class is a child of `ctypes.Structure`. This is made to write and read
    RAT headers correctly and easily preserving the size of each field in bytes.

    Parameters
    ----------
    magiclong : int
        Used for RAT version control.
    version : float
        RAT version (currently 1.0 and 2.0 versions are available).
    ndim : int
        Number of array's dimension.
    nchannel : int
        Number of channels.
    shape : list
        The shape of the data array.
    var : int
        Specifies data type according to IDL's convention. More information
        [here](http://www.exelisvis.com/docs/IDL_Data_Types.html).
    sub : int
        To be implemented.
    rattype : int
        To be implemented.

    Notes
    -----
    Due to some degree of redundancy in RatHeader (there are 3 connected attributes:
    `shape`, `ndim`, `nchannel`), the priority in determining the value of `ndim`
    and `nchannel` parameters is given to `shape` parameter. That means that `ndim`
    is calculated as `len(shape)` and `nchannel` can be given as `**kwargs`, if not
    given then it is equal either to `prod(shape[2::])` if `ndim > 2` or to `1` if
    `ndim < 2`.
    """

    _pack_ = 1
    _fields_ = [
        ("magiclong", ctypes.c_int),
        ("version", ctypes.c_float),
        ("ndim", ctypes.c_int),
        ("nchannel", ctypes.c_int),
        ("idl_shape", ctypes.c_int * 8),
        ("var", ctypes.c_int),
        ("sub", ctypes.c_int * 2),
        ("rattype", ctypes.c_int),
        ("reserved", ctypes.c_int * 9),
    ]

    def __init__(self, **kwargs):
        """
        Create an object of RatHeaderRat class.

        Create RatHeaaderRat instance. Due to redundancy of `RAT` header
        (there are 3 connected attributes: `shape`, `ndim`, `nchannel`).
        Until the latter 2 are not specified explicitly, their value is
        calculated based on `shaped`.

        To specify the data type of a `RAT` file one may use either `dtype`
        keyword with numpy data type or a string as a value, or a `var`
        keyword according to IDL's regulations.

        Parameters
        ----------
        array : numpy array
            A numpy array which is used to define/override `shape` and `dtype`.
        shape : list
            A shape of a `RAT` file.
        ndim : int
            Number of dimensions, if not given, then parsed from `shape`.
        nchannel : int
            Number of channels, if not given, then parsed from `shape`.
        var : int
            A data type according to IDL's regulation.
        dtype : numpy.dtype or a string
            A data type of a `RAT` file.
        sub : list
        rattype : int
            Type of a `RAT` file.
        """

        if 'array' in kwargs:
            kwargs['shape'] = kwargs['array'].shape
            kwargs['dtype'] = kwargs['array'].dtype

        self.magiclong = 844382546
        self.version = 2.0
        # checking kwargs
        if 'shape' in kwargs:
            # reverse the shape so it corresponds to RAT convention
            self.idl_shape = (ctypes.c_int * 8)(*kwargs['shape'][::-1])
            if 'ndim' in kwargs:
                self.ndim = ctypes.c_int(kwargs['ndim'])
            else:
                self.ndim = ctypes.c_int(len(kwargs['shape']))

            if 'nchannel' in kwargs:
                self.nchannel = ctypes.c_int(kwargs['nchannel'])
            else:
                if len(kwargs['shape']) <= 2:
                    #    self.nchannel = ctypes.c_int(0)
                    self.nchannel = ctypes.c_int(1)
                else:
                    #    self.nchannel = ctypes.c_int(kwargs['shape'][-1])
                    self.nchannel = ctypes.c_int(np.prod(kwargs['shape'][2::]))
        if 'var' in kwargs:
            self.var = ctypes.c_int(kwargs['var'])
        if 'rattype' in kwargs:
            self.rattype = ctypes.c_int(kwargs['rattype'])
        if 'sub' in kwargs:
            self.sub = (ctypes.c_int * 2)(*kwargs['sub'])
        else:
            self.sub = (ctypes.c_int * 2)(1, 1)
        if 'dtype' in kwargs:
            if type(kwargs['dtype']) == type:
                data_type = np.dtype(kwargs['dtype']).name
            elif type(kwargs['dtype'] == str):
                data_type = kwargs['dtype']
            self.var = get_var(data_type)


class RatHeaderInfo(ctypes.Structure):
    """
    Contains a 100-character line for RatFile description.

    Attributes
    ----------
    info : numpy.ndarray
        A 1-dimensional array of characters representing the RatFile description.
    """

    _pack_ = 1
    _fields_ = [("info", ctypes.c_char * 100)]


class RatHeaderGeo(ctypes.Structure):
    """
    Contains positioning information.

    Attributes
    ----------
    projection : numpy.int16
        Used for projection.
    ps_east : numpy.float64
        Easting of the polar stereographic projection.
    ps_north : numpy.float64
        Northing of the polar stereographic projection.
    min_east : numpy.float64
        Minimum easting.
    min_north : numpy.float64
        Minimum northing.
    zone : numpy.int16
        Zone information.
    hemisphere : numpy.int16
        Hemisphere information.
    long0scl : numpy.float64
        Longitude scaling.
    max_axis_ell : numpy.float64
        Maximum axis of the ellipsoid.
    min_axis_ell : numpy.float64
        Minimum axis of the ellipsoid.
    dshift_tx : numpy.float64
        Translation along the x-axis.
    dshift_ty : numpy.float64
        Translation along the y-axis.
    dshift_tz : numpy.float64
        Translation along the z-axis.
    dshift_rx : numpy.float64
        Rotation around the x-axis.
    dshift_ry : numpy.float64
        Rotation around the y-axis.
    dshift_rz : numpy.float64
        Rotation around the z-axis.
    dshift_scl : numpy.float64
        Scaling factor.
    dshift_info : numpy.ndarray
        Array of characters containing shift information.
    reserved : numpy.int8[18]
        Reserved bytes.
    """

    _pack_ = 1
    _fields_ = [
        ("projection", ctypes.c_short),
        ("ps_east", ctypes.c_double),
        ("ps_north", ctypes.c_double),
        ("min_east", ctypes.c_double),
        ("min_north", ctypes.c_double),
        ("zone", ctypes.c_short),
        ("hemisphere", ctypes.c_short),
        ("long0scl", ctypes.c_double),
        ("max_axis_ell", ctypes.c_double),
        ("min_axis_ell", ctypes.c_double),
        ("dshift_tx", ctypes.c_double),
        ("dshift_ty", ctypes.c_double),
        ("dshift_tz", ctypes.c_double),
        ("dshift_rx", ctypes.c_double),
        ("dshift_ry", ctypes.c_double),
        ("dshift_rz", ctypes.c_double),
        ("dshift_scl", ctypes.c_double),
        ("dshift_info", ctypes.c_char * 64),
        ("reserved", ctypes.c_byte * 18),
    ]


class RatHeaderEmpty(ctypes.Structure):
    """
    Empty structure with reserved bytes.

    Attributes
    ----------
    reserved : numpy.int32[25]
        Reserved bytes.
    """

    _pack_ = 1
    _fields_ = [("reserved", ctypes.c_int * 25)]


class RatHeader(ctypes.Structure):
    """
    Contains RatHeaderRat, RatHeaderInfo, and RatHeaderGeo.

    Create RatHeaaderRat instance. Due to redundancy of ``RAT`` header
    (there are 3 connected attributes: ``shape``, ``ndim``, ``nchannel``).
    Until the latter 2 are not specified explicitly, their value is
    calculated based on ``shaped``.

    To specify the data type of a ``RAT`` file one may use either ``dtype``
    keyword with numpy data type or a string as a value, or a ``var``
    keyword according to IDL's regulations.

    Parameters
    ----------
    kwargs : dict
        Keywords for initializing RatHeaderRat.

    Attributes
    ----------
    Rat : RatHeaderRat
        RatHeaderRat instance.
    Info : RatHeaderInfo
        RatHeaderInfo instance.
    Geo : RatHeaderGeo
        RatHeaderGeo instance.
    Stat : RatHeaderEmpty
        Reserved structure.
    Reserved1 : RatHeaderEmpty
        Reserved structure.
    Reserved2 : RatHeaderEmpty
        Reserved structure.
    Reserved3 : RatHeaderEmpty
        Reserved structure.
    Reserved4 : RatHeaderEmpty
        Reserved structure.
    Reserved5 : RatHeaderEmpty
        Reserved structure.
    """

    _pack_ = 1
    _fields_ = [
        ("Rat", RatHeaderRat),
        ("Info", RatHeaderInfo),
        ("Geo", RatHeaderGeo),
        ("Stat", RatHeaderEmpty),
        ("Reserved1", RatHeaderEmpty),
        ("Reserved2", RatHeaderEmpty),
        ("Reserved3", RatHeaderEmpty),
        ("Reserved4", RatHeaderEmpty),
        ("Reserved5", RatHeaderEmpty),
    ]

    def __init__(self, **kwargs):
        self.Rat = RatHeaderRat(**kwargs)


class RatFile:
    """
    Class for manipulating RAT formatted files.

    Parameters
    ----------
    filename : str
        Either a RAT filename (with *.rat extension) if the file is in the
        current working directory or an absolute path of the file.

    Attributes
    ----------
    filename : str
        The filename of the RAT file.
    Header : RatHeader
        The RAT file header.
    shape : tuple
        The shape of the numpy array.
    dtype : numpy.dtype
        The data type of the RAT file.
    ndim : int
        Number of dimensions.
    info : str
        Information about the RAT file.
    var : int
        Data type according to IDL's regulation.
    nchannel : int
        Number of channels.
    exists : bool
        Indicates whether the file exists.

    Notes
    -----
    If the file doesn't exist, the function creates a new RatFile instance
    with an empty header, not an empty file.
    """

    def __init__(self, filename):
        """
        Initialize a RAT file.

        If the file exists, reads the file's header; if not, then initializes
        the empty ``RatFile`` instance with a given filename.

        Parameters
        ----------
        filename : str
            Either a RAT filename (with *.rat extension) if the file is in the
            current working directory or an absolute path of the file.

        Returns
        -------
        RatFile instance

        Notes
        -----
        If the file doesn't exist, the function creates a new ``RatFile``
        instance with an empty header, not an empty file.
        """
        self.filename = filename
        self.Header = RatHeader()
        # the shape of numpy array
        self.shape = ()
        try:
            self.version, self.xdrflag = self.get_version()
            self.read_header()
            self.shape = self._get_shape()
            self.ndim = len(self.shape)
            self.dtype = self._get_dtype()
            self.info = self.Header.Info.info.decode()
            self.var = int(self.Header.Rat.var)
            self.nchannel = int(self.Header.Rat.nchannel)
            self.exists = True
        except (OSError, IndexError):
            self.exists = False

    @classmethod
    def _ioerror(cls, msg):
        """
        Print error message and raise OSError.

        Parameters
        ----------
        msg : str
            Error message.

        Raises
        ------
        OSError
        """
        print(red + msg + endc)
        raise OSError(msg)

    def create(self, shape=None, header=None, **kwargs):
        """
        Create an empty ``RAT`` file and write a RAT header into it.

        Create an empty ``rat`` file with given parameters and write a RAT
        header. Either ``RatHeader`` instance or ``shape`` list should be
        given as an argument. If both are given, then the value of ``shape``
        overrides the ``RatHeaderRat.shape`` value. It is necessary to specify
        a ``dtype`` keyword if the proper datatype wasn't specified in
        header's ``RatHeaderRat.var`` attribute).

        Parameters
        ----------
        shape : list, optional
            The shape of the data to store in *.rat file.
        header : RatHeaderRat, optional
            A rat header.
        kwargs : dict
            Additional keyword arguments.

        Keywords
        --------
        dtype : type or str, optional
            Data type.
        rattype : int, optional
            Specifies RAT file type.

        Returns
        -------
        None

        Raises
        ------
        IOError

        Notes
        -----
        1) The ``dtype`` keyword may be given either as a string (i.e. 'int16')
        or as a numpy ``type`` instance (i.e. ``np.int16``).
        2) The function creates an empty binary file which is a sparse file for
        Linux and Windows. OS X, however, doesn't support sparse files so it
        will be just an empty file there.
        """
        # raise an error if neither header nor shape is given as an arg
        if (header is None) and (shape is None):
            self._ioerror('Please, specify either shape or header!')

        if header is not None:
            self.Header = header

        if shape is not None:
            self.Header.Rat.idl_shape = (ctypes.c_int * 8)(*shape[::-1])

        if 'dtype' in kwargs:
            if type(kwargs['dtype']) == type:
                data_type = np.dtype(kwargs['dtype']).name
            elif type(kwargs['dtype'] == str):
                data_type = kwargs['dtype']
            self.Header.Rat.var = get_var(data_type)

        self.shape = self._get_shape()
        self.dtype = self._get_dtype()
        self.Header.Rat.ndim = ctypes.c_int(len(self.shape))
        self.Header.Rat.nchannel = ctypes.c_int(int(np.prod(self.shape[2:])))

        if 'rattype' in kwargs:
            self.Header.Rat.rattype = ctypes.c_int(kwargs['rattype'])

        # calculate the needed size of an empty file
        # n_bytes = reduce(lambda x, y: x * y, self.shape) * self.dtype.itemsize

        # write the Header and truncate the file
        with open(self.filename, 'wb') as lun:
            lun.write(self.Header)
            # lun.truncate(n_bytes)
            self.exists = True

        return self

    def write(self, arr=[], **kwargs):
        """
        Write either a whole data array or a block of data into a rat file.

        The following usages are available:

        Write a whole data array either with or without the ``header`` keyword.
        In the latter case, the parameters of the array (shape and dtype) are
        parsed into an empty ``header`` instance and written to the file. In
        this case, it's also possible to explicitly specify ``rattype`` keyword.

        When specifying the ``header`` keyword it is also possible to specify
        ``shape``, ``dtype``, and ``rattype`` keywords so the values of these
        parameters contained in ``header`` will be overwritten.

        When using a ``block`` keyword it is supposed that the header was
        already written (the entire shape of the ``RAT`` file should be known
        prior block writing).

        To write only a ``header`` one should use ``rat.write(header=header)``
        command, where ``rat`` is a ``RatFile`` instance and ``header`` is a
        ``RatHeader`` instance.

        Using a ``block`` keyword for 2D arrays the offset in both axes is
        possible (i.e. if the shape specified in the header is [100,200], then
        a block can be equal to [20,50, 130, 200]). In other cases only an
        offset along the first axis is supported. The data type of the ``arr``
        should correspond to the one given in previously written ``header``,
        otherwise, an error will be raised.

        Parameters
        ----------
        arr : numpy.ndarray, optional
            Array to be stored in rat file.
        kwargs : dict
            Additional keyword arguments.

        Keywords
        --------
        block : list, optional
            A position where to write an array; has the following form
            [start_1, stop_1, ..., ..., start_N, stop_N], where N is a number
            of array's dimensions.
        header : RatHeaderRat, optional
            RAT header.
        shape : list, optional
            Shape to overwrite the shape of the header.
        dtype : type or str, optional
            Data type to be written into the header.
        rattype : int, optional
            Specifies RAT file type.

        Raises
        ------
        IOError
        """

        arr = np.asarray(arr)
        if arr.ndim == 0:
            arr = arr.reshape(1)

        # block writing
        if 'block' in kwargs:
            # check if datatype of arr and of 'var' are the same
            self._check_dtypes(arr)

            block = kwargs['block']
            # check if the block var meets the requirements
            block = self._check_block(block, arr=arr)

            if 'header' in kwargs:
                self._ioerror(
                    'The header should have been written prior to block ' 'writing!'
                )

            # for 2D arrays an offset for both axes is allowed
            if arr.ndim == 2:
                # create zero-filled numpy array of self.shape
                # and fill it with arr.data corresponding to block array values
                temp = np.zeros([arr.shape[0], self.shape[1]], dtype=arr.dtype)
                temp[:, block[2] : block[3]] = arr
                arr = temp
                offset = 1000 + self.shape[1] * block[0] * arr.itemsize

            else:
                offset = (
                    1000 + np.ravel_multi_index(block[::2], self.shape) * arr.itemsize
                )

            with open(self.filename, 'r+') as lun:
                # The header should be written prior block writing
                lun.seek(offset)
                arr.tofile(lun)
                lun.flush()
                self.exists = True
            return

        # no block writing
        else:
            if 'header' in kwargs:
                self.Header = kwargs['header']
                # modify existing header if needed
                if 'shape' in kwargs:
                    self.Header.Rat.idl_shape = (ctypes.c_int * 8)(
                        *kwargs['shape'][::-1]
                    )
                    self.shape = tuple(kwargs['shape'])
                if 'dtype' in kwargs:
                    if type(kwargs['dtype']) == type:
                        data_type = np.dtype(kwargs['dtype']).name
                    elif type(kwargs['dtype'] == str):
                        data_type = kwargs['dtype']
                    self.Header.Rat.var = get_var(data_type)
                if 'rattype' in kwargs:
                    self.Header.Rat.rattype = kwargs['rattype']
                # check if datatypes of array and header are equal
                if arr.size > 0:
                    self._check_dtypes(arr)

            elif arr.size > 0:
                # parse array parameters to the Header
                self.Header = RatHeader(
                    shape=(arr.shape if 'shape' not in kwargs else kwargs['shape']),
                    var=np.ctypeslib.as_array(get_var(arr.dtype)),
                )
                if 'rattype' in kwargs:
                    self.Header.Rat.rattype = kwargs['rattype']

            else:
                self._ioerror('Specify a header, an array or both!')

            self.dtype = self._get_dtype()
            self.shape = self._get_shape()
            self.Header.Rat.ndim = ctypes.c_int(len(self.shape))
            self.Header.Rat.nchannel = ctypes.c_int(int(np.prod(self.shape[2:])))

            n_bytes_total = (
                1000 + reduce(lambda x, y: x * y, self.shape) * self.dtype.itemsize
            )

            with open(self.filename, 'wb') as lun:
                lun.write(self.Header)
                if arr.size > 0:
                    arr.tofile(lun)
                self.exists = True
                if lun.tell() > n_bytes_total:
                    warnings.warn(
                        "The size of the RAT file exceed! The array "
                        "is written outside the header's dimensions!"
                    )
                lun.flush()
            return

    # --------------------------------------------------------------------------
    def read(self, **kwargs):
        """
        Read either the entire data array or a block of data from a RAT file.

        Either the whole data array or a block of data can be read from the rat
        file.

        Parameters
        ----------
        block : list, optional
            A block of data to read; has the following form
            [start_1, stop_1, ..., ..., start_N, stop_N], where N is a number
            of array's dimensions.

        Returns
        -------
        arr : numpy.ndarray
            Read array.

        Raises
        ------
        IOError
        """
        if self.exists == False:
            self._ioerror('ERROR: The file "%s" does not exist' % self.filename)
        if 'block' in kwargs:
            block = kwargs['block']
            # check if the block var meets the requirements
            block = self._check_block(block)
        else:
            block = np.zeros(2 * len(self.shape), dtype=int)
            block[1::2] = self.shape

        ind = tuple(map(lambda x, y: slice(x, y, None), block[::2], block[1::2]))

        if self.version == 2.0:
            offset = 1000
        elif self.version == 1.0:
            offset = int(104 + 4 * self.Header.Rat.ndim + 4 * self.xdrflag)
        else:
            self._ioerror('ERROR: RAT version not supported')

        with open(self.filename, 'rb') as lun:
            mm = mmap.mmap(lun.fileno(), length=0, access=mmap.ACCESS_READ)

            arr = np.ndarray.__new__(
                np.ndarray, self.shape, dtype=self.dtype, buffer=mm, offset=offset
            )[ind]
            if self.xdrflag == 1:
                arr = arr.byteswap()
        arr_new = np.zeros(shape=arr.shape, dtype=arr.dtype)
        arr_new[:] = arr
        return arr_new

    # --------------------------------------------------------------------------
    def mread(self, **kwargs):
        """
        Read data from a RAT file as a NumPy array using memory map.

        Works with both RAT 1.0 and 2.0 files. Allows reading data in blocks
        along all axes.

        Parameters
        ----------
        block : list, optional
            The block of data to read; has the following form
            [start_1, stop_1, ..., ..., start_N, stop_N], where N is a number
            of axes.

        Returns
        -------
        arr : numpy.ndarray
            Read array.

        Raises
        ------
        IOError
            If the file doesn't exist, if RAT version is not recognized,
            and when the block doesn't correspond to the shape of the RAT header.
        """
        if self.exists == False:
            self._ioerror('ERROR: The file is not found')
        if 'block' in kwargs:
            block = kwargs['block']
            # check if the block var meets the requirements
            block = self._check_block(block)
        else:
            block = np.zeros(2 * len(self.shape), dtype=int)
            block[1::2] = self.shape

        ind = tuple(map(lambda x, y: slice(x, y, None), block[::2], block[1::2]))

        if self.version == 2.0:
            offset = 1000
        elif self.version == 1.0:
            offset = int(104 + 4 * self.Header.Rat.ndim + 4 * self.xdrflag)
        else:
            self._ioerror('ERROR: RAT version not supported')

        with open(self.filename, 'rb') as lun:
            mm = mmap.mmap(lun.fileno(), length=0, access=mmap.ACCESS_READ)

            arr = np.ndarray.__new__(
                np.ndarray, self.shape, dtype=self.dtype, buffer=mm, offset=offset
            )[ind]
            if self.xdrflag == 1:
                arr = arr.byteswap()
        return arr

    def append(self, arr):
        """
        Append the RAT file with a given array along the first axis.

        Parameters
        ----------
        arr : numpy.ndarray
            The array to write.

        Raises
        ------
        Warning
            If the dimensions specified in the file's header are exceeded.
        """
        # self.read_header()
        # check if datatype of arr and of 'var' are the same
        self._check_dtypes(arr)

        if arr.ndim == len(self.shape) - 1:
            arr = arr[np.newaxis, ...]

        # check if the array's and header's shape correspond to each other
        if len(self.shape) > 1 and (self.shape[1:] != arr.shape[1:]):
            self._ioerror(
                'The shape specified in the header {} and the shape of '
                'the array {} don\'t correspond to each other!'.format(
                    str(self.shape[1:]), str(arr.shape[1:])
                )
            )

        n_bytes_total = 1000 + reduce(lambda x, y: x * y, self.shape) * arr.itemsize

        with open(self.filename, 'r+b') as lun:
            lun.seek(0, 2)
            arr.tofile(lun)
            if lun.tell() > n_bytes_total:
                warnings.warn(
                    "The size of the RAT file exceed! The array is "
                    "written outside the header's dimensions!"
                )

        return

    # --------------------------------------------------------------------------
    def read_header(self):
        """
        Read RAT file header, supporting both RAT 1.0 and 2.0 versions.

        Raises
        ------
        IOError
            If RAT version is not supported.
        """
        # reading the version
        self.version = self.get_version()[0]
        """Read ``RAT`` header; supports both ``RAT`` 1.0 and 2.0 versions."""
        if self.version == 2.0:
            with open(self.filename, 'rb') as lun:
                lun.readinto(self.Header)

        elif self.version == 1.0:
            warnings.warn('Old RAT v1.0 format!')
            if self.xdrflag == 1:
                data_type = '>i4'
                offset = 4 * 4
            else:
                data_type = '<i4'
                offset = 3 * 4

            with open(self.filename, 'rb') as lun:
                ndim = int(np.fromfile(file=lun, dtype=data_type, count=1))
                shape = np.fromfile(file=lun, dtype=data_type, count=ndim).tolist()
                shape = shape[::-1]
                var = int(np.fromfile(file=lun, dtype=data_type, count=1))
                rattype = int(np.fromfile(file=lun, dtype=data_type, count=1))
                lun.seek(offset, 1)
                info = np.fromfile(file=lun, dtype="B", count=80).tostring().rstrip()

            # initialize the header
            self.Header = RatHeader(shape=shape, ndim=ndim, var=var, rattype=rattype)
            self.Header.Info.info = info

        else:
            self._ioerror('ERROR: RAT version not supported')

        self.dtype = self._get_dtype()

    # --------------------------------------------------------------------------

    def write_envi_header(self, info='', sensorType='DLR F-SAR'):
        """
        Write ENVI header for the RAT file.

        Parameters
        ----------
        info : str, optional
            Description to be included in the ENVI header.
        sensorType : str, optional
            Sensor type information.

        Raises
        ------
        None
        """
        hdrFile = self.filename + '.hdr'
        with open(hdrFile, 'w') as f:
            f.write('ENVI\n')
            if len(info) > 0:
                f.write('description     = {%s}\n' % info)
            f.write('samples         = %i\n' % self.Header.Rat.idl_shape[0])
            f.write('lines           = %i\n' % self.Header.Rat.idl_shape[1])
            f.write('bands           = %i\n' % np.maximum(self.Header.Rat.nchannel, 1))
            f.write('header offset   = 1000\n')
            f.write('file type       = ENVI Standard\n')
            f.write('data type       = %i\n' % self.Header.Rat.var)
            f.write('interleave      = bsq\n')
            f.write('byte order      = 0\n')
            if len(sensorType) > 0:
                f.write('sensor type     = %s\n' % sensorType)

    def write_geo_envi_header(self):
        """
        Write geographic information to the ENVI header.

        Raises
        ------
        None
        """
        # tmpl = Template(resource_string(__package__ + '.templates', 'envihdr.tpl'))
        tmpl = Template(
            resources.files(__package__).joinpath('.templates', 'envihdr.tpl')
        )
        envi_hdr = tmpl.render(file=self.filename, hdr=self.Header)
        with open(self.filename + '.hdr', 'w') as f:
            f.write(envi_hdr)

    # --------------------------------------------------------------------------

    def help(self):
        """
        Print basic information about the RAT file.

        Print shape, data type, and info from the file's header.
        If the file doesn't exist, then print "empty file".
        """
        print()
        print("FILE  : ", os.path.abspath(self.filename))
        if self.exists == True:
            print("SHAPE : ", self.shape)
            print("TYPE  : ", self._get_dtype())
            print("INFO  : ", self.Header.Info.info)
        else:
            print("--- empty file ---")

    # --------------------------------------------------------------------------

    def get_version(self):
        """
        Get the version of RAT file: 1.0 or 2.0.

        Returns
        -------
        tuple
            A tuple containing the RAT file version (1.0 or 2.0) and XDR flag.

        Notes
        -----
        This method reads the magic number and checks whether it corresponds to
        RAT version 1.0 or 2.0. If the magic number indicates version 1.0, it
        additionally checks for endianness using the number of dimensions (ndim).

        If a RAT V1 file is detected, it returns a tuple (1.0, 0). If a RAT V2
        file is found, it returns a tuple (2.0, 0).

        Raises
        ------
        OSError
            If an error occurs during file reading or if the RAT version is not recognized.
        """
        Header = RatHeader()
        magiclong = Header.Rat.magiclong
        magicreal = 0

        with open(self.filename, 'rb') as lun:
            magicreal = np.fromfile(file=lun, dtype="i4", count=1)

        if magicreal != magiclong:  # Check if maybe we have a RAT V1 File...
            with open(self.filename, 'rb') as lun:
                ndim = np.fromfile(file=lun, dtype="<i4", count=1)

            xdrflag = 0
            if ndim < 0 or ndim > 9:
                ndim = ndim.byteswap()
                xdrflag = 1

            if ndim < 0 or ndim > 9:
                print(red + "ERROR: format not recognised!" + endc)
                return False, False

            version = 1.0
        else:  # -------------- Yeah, RAT 2.0 found
            with open(self.filename, 'rb') as lun:
                lun.seek(4)
                version = np.fromfile(file=lun, dtype="float32", count=1)[0]

            xdrflag = 0

        return version, xdrflag

    def _check_block(self, block, **kwargs):
        """
        Check the validity of the block for writing an array.

        Parameters
        ----------
        block : list
            A position where to write an array, in the form [start_1, stop_1, ..., start_N, stop_N].
        kwargs : dict
            Keyword arguments.

        Returns
        -------
        block : np.ndarray
            Validated block.

        Raises
        ------
        IOError
            If block extent is not given by integers, contains negative values,
            or exceeds the array shape.
        """
        if len(block) == 4:  # only 2D block provided
            block = list(block)
            dimlist = list(self.shape)
            dimlist[dimlist.index(max(dimlist))] = 0
            dimlist[dimlist.index(max(dimlist))] = 0
            for k, dim in enumerate(dimlist):
                if dim != 0:
                    block.insert(k * 2, dim)
                    block.insert(k * 2, 0)

        stop_more_than_shape = reduce(
            lambda x, y: x or y, (map(lambda x, y: (x > y), block[1::2], self.shape))
        )
        if stop_more_than_shape:
            self._ioerror('Value of block exceeds the array shape!')

        block = np.asarray(block)
        if block.dtype.kind not in ('i', 'u'):
            self._ioerror('Block extent must be given by integers!')

        if np.min(block) < 0:
            self._ioerror('The items in block must be nonnegative!')

        if 'arr' in kwargs:
            if len(block) // 2 != kwargs['arr'].ndim:
                self._ioerror(
                    'The dimensions of block do not correspond to the '
                    'dimensions of array!'
                )

            block_not_shape = reduce(
                lambda x, y: x or y,
                map(
                    lambda x, y, z: (x - y) != z,
                    block[1::2],
                    block[::2],
                    kwargs['arr'].shape,
                ),
            )
            if block_not_shape:
                self._ioerror(
                    'Length of block components {} does not correspond to'
                    ' the shape of the array {}!'.format(
                        str(block), str(kwargs['arr'].shape)
                    )
                )
        return block

    def _check_dtypes(self, arr):
        """
        Check if dtypes of given array and the one in header are equal.

        Parameters
        ----------
        arr : np.ndarray
            The array to be checked.

        Raises
        ------
        IOError
            If data type of the array to be written does not correspond
            to the one specified in the file's header.
        """
        if self.Header.Rat.var != get_var(arr.dtype).value:
            self._ioerror(
                'The data type of the array to be written and '
                'the one specified in file\'s header '
                'don\'t correspond to each other!'
            )

    def _get_dtype(self):
        """
        Get data type given by `Header.Rat.var`.

        Returns
        -------
        dtype : np.dtype
            Data type.
        """
        try:
            return np.dtype(dtype_dict[self.Header.Rat.var])
        except KeyError:
            self._ioerror('The data type is either not specified or ' 'not supported!')

    def _get_shape(self):
        """
        Get numpy array shape from `Header.Rat.shape` that is IDL style.

        Returns
        -------
        shape : tuple
            Numpy array shape.
        """
        shape = np.ctypeslib.as_array(self.Header.Rat.idl_shape)
        shape = shape[shape != 0]
        return tuple(shape[::-1])


def check_ratformat(filename):
    """
    Check if the RAT file format is valid.

    Parameters
    ----------
    filename : str
        The path to the RAT file.

    Returns
    -------
    bool
        Returns True if the file format is RAT2 (version 2.0), and False otherwise.

    Notes
    -----
    This function reads the first 4 bytes (magic number) of the RAT file to
    determine if it follows the RAT2 (version 2.0) format.

    Raises
    ------
    OSError
        If an error occurs during file reading.
    """
    with open(filename, 'rb') as lun:
        magiclong = lun.read(4)
        if magiclong == b'RAT2':
            return True
        else:
            return False


def get_var(dtype):
    """
    Get ``RatHeaderRat.var`` value given ``dtype``.

    Parameters
    ----------
    dtype : numpy.dtype
        The NumPy data type.

    Returns
    -------
    ctypes.c_int
        The value corresponding to ``RatHeaderRat.var`` for the given data type.

    Raises
    ------
    OSError
        If the data type is not supported according to the ``dtype_dict`` mapping.
    """
    var = [key for (key, value) in dtype_dict.items() if value == dtype]
    if var == []:
        RatFile._ioerror('The data type is not supported!')
    else:
        return ctypes.c_int(var[0])


# data type dictionary to net RAT's and np's data formats
dtype_dict = {
    1: 'uint8',
    2: 'int16',
    3: 'int32',
    4: 'float32',
    5: 'float64',
    6: 'complex64',
    9: 'complex128',
    12: 'uint16',
    13: 'uint32',
    14: 'int64',
    15: 'uint64',
}


class Py2Xml:
    """
    Class to read/write DLR-STE XML documents (primarily used by the STEP processor for processing parameters)

    Allows python object-like access to parameters.

    For example, for STEP processing parameters (XML files in RGI/RGI-RDP):

    pp = Py2Xml('path_to_pp.xml')
    print(pp.v0)     # v0 is automatically converted to floating point
    print(pp.r[:10]) # first 10 values of range vector, also floating point

    The parameter object also allows dictionary-like access to handle problematic parameter names
    (which clash with python keywords). For example:

    print(pp['lambda']) # pp.lambda would be a syntax error
    print(pp['pass'])   # same as above
    print(pp['v0'])     # dictionary style access works for other parameters, too!

    The class provides full read/write support. Parameter values are changed by standard assignment
    and structures can be saved using the write method:

    pp.v0 = 100
    pp.write('path_to_new_pp.xml')

    Parameters
    ----------
    root : str or Element
        The path to the XML file or an Element object representing the root.

    Raises
    ------
    ValueError
        If the expected "object" element is not found below the root element.

    Methods
    -------
    copy()
        Create a deep copy of the Py2Xml instance.

    params()
        Get a list of parameter names.

    write(filename)
        Write the XML structure to the specified file.

    tostring()
        Get the XML structure as a string.

    fromstring(string)
        Create a Py2Xml instance from an XML string.

    Attributes
    ----------
    __iterend__ : bool
        Internal attribute to track the end of iteration.

    Notes
    -----
    The class uses ElementTree for XML parsing and manipulation. It provides methods to convert
    between XML and Python objects, allowing convenient access and modification of parameters.

    See Also
    --------
    xml.etree.ElementTree
    """

    def __init__(self, root):
        """
        Initialize Py2Xml instance.

        Parameters
        ----------
        root : str or Element
            The path to the XML file or an Element object representing the root.

        Raises
        ------
        ValueError
            If the expected "object" element is not found below the root element.
        """
        if isinstance(root, str):
            self.__dict__['__root__'] = ET.parse(root).find('object')
        else:
            self.__dict__['__root__'] = root

        if self.__root__ is None:
            raise ValueError('Expected an "object" element below the root element!')

        self.__dict__['__iterend__'] = False

    def __getstate__(self):
        return ET.tostring(self.__root__, encoding=str)

    def __setstate__(self, root):
        self.__dict__['__root__'] = ET.fromstring(root)
        self.__dict__['__iterend__'] = False

    def copy(self):
        return Py2Xml(copy.deepcopy(self.__dict__['__root__']))

    def __getparam__(self, name):
        p = [
            p
            for p in self.__root__.iterchildren('parameter')
            if p.attrib['name'] == name
        ]
        if len(p) != 1:
            raise AttributeError(
                'Expected a unique match parameter name "%s", got %i matches.'
                % (name, len(p))
            )

        return [p[0].find(tag) for tag in ('remark', 'datatype', 'value')]

    def params(self):
        return [p.attrib['name'] for p in self.__root__.iterchildren('parameter')]

    @staticmethod
    def xml2val(v, t):
        type = t.text
        shape = t.attrib['length']
        shape = np.asarray([np.uint64(d) for d in shape.split()])[::-1]
        size = np.prod(shape)

        if type == 'pointer':
            p = v.find('parameter')
            return Py2Xml.xml2val(*[p.find(t) for t in ('value', 'datatype')])

        if type == 'struct':
            obj_arr = [Py2Xml(obj) for obj in v.iterchildren('object')]
            return obj_arr[0] if size <= 1 else obj_arr

        conv = {
            'int': int,
            'long': int,
            'float': np.float,
            'double': np.double,
            'string': lambda s: s,
        }
        try:
            if size > 1:
                val = np.asarray(
                    [conv[type](v) for v in v.text.strip('[]').split(',')]
                ).reshape(shape)
            else:
                val = conv[type](v.text)
        except KeyError:
            print(
                'PY2XML WARNING: Unsupported data type "%s" encountered. Skipping!'
                % (type)
            )
            return None

        return val

    @staticmethod
    def val2xml(v, t, value):
        cdict = {
            str: (str, 'string'),
            int: (str, 'long'),
            float: (str, 'double'),
            complex: (lambda z: f'({z.real},{z.imag})', 'complex'),
        }
        cdict[np.uint8] = cdict[int]
        cdict[np.int32] = cdict[int]
        cdict[np.uint32] = cdict[int]
        cdict[np.int64] = cdict[int]
        cdict[np.uint64] = cdict[int]
        cdict[np.float32] = (str, 'float')
        cdict[np.float64] = cdict[float]
        cdict[np.complex64] = cdict[complex]
        cdict[bool] = cdict[str]

        if t.text == 'pointer':
            p = v.find('parameter')
            return Py2Xml.val2xml(
                *([p.find(t) for t in ('value', 'datatype')] + [value])
            )

        try:
            vsize = 1 if isinstance(value, str) else len(value)
        except TypeError:
            vsize = 1

        t.attrib['length'] = str(vsize)
        v.clear()
        if vsize == 1 and not isinstance(value, Py2Xml):
            t.text = cdict[type(value)][1]
            v.text = cdict[type(value)][0](value)
        elif all([isinstance(v, Py2Xml) for v in value]):
            t.text = 'struct'
            for obj in value:
                v.append(copy.deepcopy(obj.__root__))
        else:
            if isinstance(value, np.ndarray):
                t.attrib['length'] = ' '.join([str(l) for l in value.shape[::-1]])
                value = value.flat
            vtype = type(value[0])
            t.text = cdict[vtype][1]
            v.text = '[' + ', '.join([cdict[vtype][0](val) for val in value]) + ']'

    def __getattr__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if key == 0:
            return self
        r, t, v = self.__getparam__(key)
        return Py2Xml.xml2val(v, t)

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value
            return
        r, t, v = self.__getparam__(name)
        Py2Xml.val2xml(v, t, value)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __contains__(self, key):
        try:
            _ = self.__getparam__(key)
        except AttributeError:
            return False
        return True

    def __len__(self):
        return 1

    def __iter__(self):
        self.__iterend__ = False
        return self

    def __next__(self):
        if self.__iterend__:
            raise StopIteration()
        self.__iterend__ = True
        return self

    def update(self, obj):
        if isinstance(obj, Py2Xml):
            d = {k: obj[k] for k in obj.params()}
        else:
            try:
                d = obj.__dict__
            except AttributeError:
                d = obj
            if not isinstance(d, dict):
                raise ValueError(
                    'Expected a dictionary or an object with a __dict__ attribute!'
                )

        for k in d:
            try:
                self.__setattr__(k, d[k])
            except AttributeError:
                pass

    def __totree(self):
        ste_root = ET.Element('stexml')
        ste_root.text = '\n'
        ste_root.append(copy.deepcopy(self.__root__))
        ste_root.addprevious(
            ET.PI('xml-stylesheet', 'type="text/xsl" href="stexml.xsl"')
        )
        tree = ET.ElementTree(ste_root)
        return tree

    def write(self, filename):
        self.__totree().write(
            filename, pretty_print=True, encoding='UTF-8', xml_declaration=True
        )

    def tostring(self):
        return ET.tostring(self.__totree().getroot(), encoding='UTF-8')

    @staticmethod
    def fromstring(string):
        return Xml2Py(ET.fromstring(string).find('object'))


# for backwards-compatiblity (the old Xml2Py class is now obsolete!)
Xml2Py = Py2Xml



# flake8: noqa

# This minimal dataset was available as part of
# pysarpro 0.15 and will be retained until
# further notice.
# Testing data and additional datasets should only
# be made available by pooch
legacy_datasets = [
    'astronaut.png',
    '01-sar/signal1_ac.npy',
    '01-sar/signal1_rc.npy',
    '01-sar/signal2_ac.npy',
    '01-sar/signal2_rc.npy',
]

# Registry of datafiles that can be downloaded along with their SHA256 hashes
# To generate the SHA256 hash, use the command
# openssl sha256 filename
registry = {
    "data/astronaut.png": "88431cd9653ccd539741b555fb0a46b61558b301d4110412b5bc28b5e3ea6cb5",
    "data/01-sar/raw-img.rat": "88316eb411b3099575cb6a813911af99e61d1116486e11c75a1327a0da2b14a2",
    "data/01-sar/signal1_ac.npy": "814e847e979c6ae0c9cda290c7f6b2d7ed2bf212cf8b454a8671abf1f9d5df52",
    "data/01-sar/signal1_rc.npy": "30f43d643b162eda34061b72bf64bc9a4e1893cfdb309e1e577fa9ba19e629af",
    "data/01-sar/signal2_ac.npy": "ed62d3d0e8a7c52f6b031493f537304c77b7fd833dd8820578cb847bbcc42d60",
    "data/01-sar/signal2_rc.npy": "8c3b01775eebfcb1795aaaab9396928c3251c21d9cfe3b0484ce0e9624640138",
    "data/polsar/slc_16afrisr0107_Lhh_tcal_test.rat": "d1b50ede4c0f981ffa32beaf3d355ac2655b0a076e7e48c029c6facb3cea9845",
    "data/polsar/slc_16afrisr0107_Lhh_tcal_test.rat.hdr": "b9dbca4c6f86d14743441782dd09e12127da8a82cb0fea496f2fe3f3654e0804",
    "data/polsar/slc_16afrisr0107_Lhv_tcal_test.rat": "f045e8251d9a38e4e35a303c9929545fc115b71da57176277abffd3d99ba5683",
    "data/polsar/slc_16afrisr0107_Lhv_tcal_test.rat.hdr": "be45012e118e8b5c5a97ce1d43e10208fd5670dd323be023ea59e5e4bf2faaae",
    "data/polsar/slc_16afrisr0107_Lvh_tcal_test.rat": "8304975ca3e523afa4d3f14e5502fa9a647f180b3bbc853dda2740851cfa3f04",
    "data/polsar/slc_16afrisr0107_Lvh_tcal_test.rat.hdr": "ba19984e02c3c13eefa575cc21f96f833082901c38969df051f2ff47c70f34a7",
    "data/polsar/slc_16afrisr0107_Lvv_tcal_test.rat": "5305c2047ef648b9ff3f28547c63472a020429739ab6d8f8be2480cae08e54e2",
    "data/polsar/slc_16afrisr0107_Lvv_tcal_test.rat.hdr": "860bf405b0ce68530f302deea7824d7812cb2e554f904d4de9292b1919d88a25",
    "data/insar/common_profile.npy": "3d7b425801ca4bb55dbd7e994b644317c14570a4ebe20d6c9eeb34dea283119a",
    "data/insar/flat_earth_Mondah_S_2015_11_11_cut.npy": "dc92d9fc3f2d5d68819b4621b0df9c5f92f1e734a8b0e1f765494e4cf71ea12e",
    "data/insar/flat_earth_Mondah_S_2015_11_11_cut.rat": "8e2186e30bca1b275b95d7b9401590ede520085feac8aa6c5f73dbdb1143cdf5",
    "data/insar/image1.rat": "cb8bb55973f85604df57f077c0525e2c529b9832924bcb7115132deedede0d15",
    "data/insar/image2.rat": "7b071c7d69cf6a294b9b8b7187fe4aaef71798f55b4c47c4540e278a71d9449d",
    "data/insar/master_HH_Mondah_S_2015_11_11_cut.npy": "c7521b9fbbe560a08d539bf88d2c42bca58a84061e1ece712b918399c884d00c",
    "data/insar/master_HH_Mondah_S_2015_11_11_cut.rat": "983266a183e44e9b2cee0fa0b445b1d7258b595e0318d29c0a42df19b546b940",
    "data/insar/RH100_Mondah_S_2015_11_11_cut.npy": "f432a757aa1f5759a5b04450a46f6ac2544106ca1c5559483d6386e147b8abcc",
    "data/insar/RH100_Mondah_S_2015_11_11_cut.rat": "ad3d5a08a4d4e38e8904d396c59581a611e2d29d0ac590171dc61ea2018ade71",
    "data/insar/slave_HH_Mondah_S_2015_11_11_cut.npy": "9eaf6d91fd73be12ab310910c7ccbf5fb66651e52d8b976ca69df6aa4198315f",
    "data/insar/slave_HH_Mondah_S_2015_11_11_cut.rat": "b06a4f07197359b47c68a101e17c77e3371e24eaa91854f26da540188148015e",
    "data/polinsar/gamma_max.rat": "ef51aa6486459d64b86fca63cd0a73623a1600945753fa0bc120d10f4de28c46",
    "data/polinsar/gamma_min.rat": "5f86fc60ed7153502c2c971fe359725da091807204aa9d46d57c6443f0384c51",
    "data/polinsar/incidence_15tmpsar0302_L_t01.rat": "8072888fa7ada41ff18317e6de30e2a56d78e0c5665cb087be66f50768d57de6",
    "data/polinsar/kz_2d_demc_15tmpsar0302_15tmpsar0303_t01.rat": "8ad3a16da3ef2c621a47acafceed5d7ac07f5643b53c47d8f94cd1a608d3adb3",
    "data/polinsar/Lida_r1503.rat": "c57bda37ebf97210e16163ac487bb00d5b50fb7d61f2192280c3bc442d524b97",
    "data/polinsar/local_slope_15tmpsar0302_L_t01.rat": "f1326a8f191cd78af3e89f9e29d353449d8c9c16f44f405b09066e69b58f07d8",
    "data/polinsar/pha_flat_15tmpsar0302_15tmpsar0303_Lhh_t01.rat": "913a75e680d56d04f4fcd2f5e0dea0786b7236d3bd749c50330719f333a4fea4",
    "data/polinsar/pha_max.rat": "4bd6d30f78b5164ab8a39a31eb88e3fad847695f20851c32cb67cb498b44229e",
    "data/polinsar/pha_min.rat": "bcb3631abaa1898dd0322822544da9b245638769c8caa0f86940145a657415b1",
    "data/polinsar/slc_15tmpsar0302_Lhh_t01.rat": "aef646a151f897d49ff9f68bfe0caf504db431eb1ba1430a4a9874647f52506a",
    "data/polinsar/slc_15tmpsar0302_Lhv_t01.rat": "fdd503474723f678e967537b4155e99a4a7a8d2019b8a1591fec5d87a794cadc",
    "data/polinsar/slc_15tmpsar0302_Lvh_t01.rat": "8a5d247404d108f03fdd94bc417bc8f9d0f209f62f09bcd5e3b41f9a31ade412",
    "data/polinsar/slc_15tmpsar0302_Lvv_t01.rat": "1a6a38cd0ff2d313c02c812f48ce800f14b5caf1ceb462554dbd9ce93354933d",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lhh_t01.rat": "53dc085de9dd6b828248e54aff96d6c3dc47dd87c5ea196ba36f62f711b8ee58",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lhv_t01.rat": "4bd6b729c01b521fbea3d593828d894aff6d8e130d82fad3fadc2497d441ce01",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lvh_t01.rat": "7851e089feb307654bb912609ee8d2fe88e274bec0fe31f64693c59d732e8923",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lvv_t01.rat": "03121641edd8f48d355a310c41178f01bb92e35e76c684853c993d6db80d27bd",
    "data/tomosar/dem.rat": "226f6806773b1759ddb696b25dabb44ff776bcf727e3a33edc9f3cf41a02b229",
    "data/tomosar/kz_00.rat": "f915c9980d71c389aa9d850041d08cea8eff2fdc3443cd72301912acf5513b4b",
    "data/tomosar/kz_01.rat": "95b6be70d5e79dcd87bce1762c3f86e0131327625ee47242d186a9c440cc4039",
    "data/tomosar/kz_02.rat": "daab75242a800e6856d51bda9fe787a2247fa8a065c05b0da1600619a2dc5d57",
    "data/tomosar/kz_03.rat": "1a0ebde5c5b5d2709f05519de0a8b55bbf8589f4ae3da07b7e0f38a962531e3d",
    "data/tomosar/kz_04.rat": "7bbcd6d997cfe93befb05b9b33352030ba6d4a37bd7a508e38ca7028043c1bd1",
    "data/tomosar/kz_05.rat": "ad2dbd76e1ce7b55fdfe441714f86a4576d4166ee607f0e6594dbb1ac3a0dbd2",
    "data/tomosar/kz_06.rat": "0031fd2d43027a770c1052fd843e734beebe58899ac99656152e7f1fe229a653",
    "data/tomosar/kz_07.rat": "a26996d35359b1ed3a72d89abde140d1999939f8694705962e7e8fbe57a88381",
    "data/tomosar/kz_08.rat": "9c5e8cb255412d79059f464ac3b658a1eb6ff9c8136b8fca4761a98712071d45",
    "data/tomosar/kz_09.rat": "768af65c80338384ce881770004e016e6be4f544912ff98cbd9e209a01eabcd4",
    "data/tomosar/kz_10.rat": "989d57637ff9eb448de82de7419f1506bcaa3a4958ca2c77a4c2d11c127ae61d",
    "data/tomosar/lidar_chm.rat": "96f4e56ca8e3d20f2868bae6d77a4075139d2f26d1609f1281e8f68db8b0b15b",
    "data/tomosar/lidar_dtm.rat": "c769673c01ff6be88f8ffcb6e5cb870ad5b9ec7e70a5f3b265cbff9e5b2e577f",
    "data/tomosar/pha_dem_00.rat": "380e0633010eeda664d1cf52e09d57fb1660fe84edd79110190a09368fd9cca6",
    "data/tomosar/pha_dem_01.rat": "57db905f21e7e6d4c1b6cb13ffb6be5aacddf54b6912cc084f90bb35fac4b371",
    "data/tomosar/pha_dem_02.rat": "b6fc41ee2ed666b54d90b48890c6a36ca8347e1ab3e053473ce6e57520442753",
    "data/tomosar/pha_dem_03.rat": "b69aa4889c3c2a4cbef817c0acf556242bfbbeb0820eac038608c3c99f913a36",
    "data/tomosar/pha_dem_04.rat": "a6bdffa10c88b3fd1fa2f4562e3e5ac9de11abcd8ec6988636a0e429e57a2f6b",
    "data/tomosar/pha_dem_05.rat": "d253cdbb3eb8553ddba81339c8f74f994d7d0955df6b290e85f93118456043aa",
    "data/tomosar/pha_dem_06.rat": "ba0d99a6d608c84987491fdc20ecd790a43a336b1625062baa178bc6325d8c3a",
    "data/tomosar/pha_dem_07.rat": "a6e061b0e8e7d12629cf9327bb9c161e0cb85eaf75369176f6298aeb445e9605",
    "data/tomosar/pha_dem_08.rat": "43dc7d19985e4e5c8b3d5dd32c4e3e31cb363ca88be831d3dd7a7f2d68c62883",
    "data/tomosar/pha_dem_09.rat": "2581da4e098f57de2620015d2b9cdf74bba192032ce07a43b5dee1923f695f27",
    "data/tomosar/pha_dem_10.rat": "672245b9cb33014ae139b7be13775e689cd930c240013e55cd55498e0657ddce",
    "data/tomosar/pos_scatters_ground.npy": "a1902c3b0c428587ce1043a370852a2eb80d82eef4b969262b9675562d98c05e",
    "data/tomosar/pos_scatters_layer1.npy": "9ec4bb41928f880eb8eac2697e37732e5c853220e1f3def54ac07caf8b277dfd",
    "data/tomosar/slc_00_HH.rat": "b434350d2cf9c8c1232af6ef18b474ce4683ca1d84608b33f759ed6f3e872331",
    "data/tomosar/slc_00_HV.rat": "a32cc617ca51a8dda50b19a8d67f08d99cd6c77a3fab356cfef9c7c3d3ee81a7",
    "data/tomosar/slc_00_VV.rat": "b37ff61e689b3abe0bbf83c323b524e4ad7e40b3ae8d69d6e97602f2f4aebfad",
    "data/tomosar/slc_01_HH.rat": "bb06db40651aa757e281332c2f290d87593e79a98cd1c6c3a62f22f38a375015",
    "data/tomosar/slc_01_HV.rat": "0978b94b23ba36af92e4e1834263c1001daa5f6111d3fe6e761bf1aa7fb1a464",
    "data/tomosar/slc_01_VV.rat": "6eec3ecb5204dbc8a9bb0781daba7331d50dae18ea953f373e437be7054f9a32",
    "data/tomosar/slc_02_HH.rat": "cb57062d39f0f008d1a172b8c3f8ce311c9140d8e464e395d75e219755bd095a",
    "data/tomosar/slc_02_HV.rat": "f6e8c065f09030e4c733ae8128c4337a737d64a9da76ce422c5e45cf0e74f53f",
    "data/tomosar/slc_02_VV.rat": "33bbe3e0480891e309bb42c2b0407690fbf536178477d12ff18101168bf1b772",
    "data/tomosar/slc_03_HH.rat": "19c488eeffae4496f5bbea2e26df214a59b5f0ae7bcbed674d9db9dfdcc03184",
    "data/tomosar/slc_03_HV.rat": "bd11da7f59fdc5afb59b97901cd08db411ec927ebe7d42cc7da28b8ea6faa608",
    "data/tomosar/slc_03_VV.rat": "6f648dac736b541cc53023a9ec635275eb5ff40eecc28ea0f69fd373a7c8ed13",
    "data/tomosar/slc_04_HH.rat": "5ed7c8d8c9710779f32c8530ed1ff2151cb209693a75afeff3bb416da12f0a55",
    "data/tomosar/slc_04_HV.rat": "7d1a1267a010e00a60a7aae595155ccb6c3e80069689fd1cb0f999d60fc01778",
    "data/tomosar/slc_04_VV.rat": "539bfe527f5452d7e44f11192dd13e9d9d2387536ed05aea250f4188bddc8495",
    "data/tomosar/slc_05_HH.rat": "ce28edb56fe6affbe042168b83bee5c7bb4d6d2e671490cd91a7ddd3fa8fc7a5",
    "data/tomosar/slc_05_HV.rat": "24ebea0807e228dc4aaff1eed8346e77f79c7249a5680d5c8223aae0eef913a4",
    "data/tomosar/slc_05_VV.rat": "7959950f0b673becce3a47aded871817a087093220554508bbc7995eda3fcf3c",
    "data/tomosar/slc_06_HH.rat": "d679258e5324ee37a7400728b3f75a508590b5bf86b75df4bb9b02731c6c4ee6",
    "data/tomosar/slc_06_HV.rat": "2f3f173af1f1b5d3b45bfeb6c411b80a96b6d0e90a205358f7c52a43a524d3a6",
    "data/tomosar/slc_06_VV.rat": "cd20797422ce0eb0e09270e77b9a2d0ac96c80ef9f08c237373de865e01a6571",
    "data/tomosar/slc_07_HH.rat": "fcd7bb701e68749f6844f73fa34a6321cbb0b362518d7bcb376a4d31c1ef3d7e",
    "data/tomosar/slc_07_HV.rat": "332a96a55951b703c649719b178e5c46690712f31d8052c85a1d7e0fe67077fe",
    "data/tomosar/slc_07_VV.rat": "703897c7ed72dc698b8ff58ce26674318a453847a2b1d5218e26d527c5827a4e",
    "data/tomosar/slc_08_HH.rat": "7a6e9cc764888b0df3d2a61d381c7096f5f87b11070dd450270849108a9bfd2e",
    "data/tomosar/slc_08_HV.rat": "50f08a8e20ff55862f10938bbc54dc947679d08933685b2f6c63b3e42dcdeeeb",
    "data/tomosar/slc_08_VV.rat": "d172c4a5d03b71fb2626b8b49245f93f2bf22c22509513014a210bdb4db61ea5",
    "data/tomosar/slc_09_HH.rat": "e122bb4b9418d93153cb8aec40c70061abafa0d7ed9cab839931b47f50d3d3a0",
    "data/tomosar/slc_09_HV.rat": "40057ad2b49492c318d7eb3f5edbd47594a611dcfba5d3f4a38b6f3535233a70",
    "data/tomosar/slc_09_VV.rat": "e381df66217995d7a79702eea4082ae3ba960ba761780ba35e755a4df06c03da",
    "data/tomosar/slc_10_HH.rat": "bd34eef79940c00ccafc80f637c6d979f90b79d1ee301d6327b59233f320dc9a",
    "data/tomosar/slc_10_HV.rat": "b435efcf0366f3bcf62ac57c81b371e60879edfac001ef2fad0452ce3200f24b",
    "data/tomosar/slc_10_VV.rat": "dac25b527e7e223861311a5fa81f18919ac9d9f9c77d1c1753f41d1070dd9c3f",
    "data/DInSAR.zip": "14692fb707283319e645bda92a6924351528c149d731d2fbc63c72cf3e445f1c",

}

registry_urls = {
    "data/01-sar/raw-img.rat": "https://zenodo.org/records/10797182/files/raw-img.rat",
    "data/polsar/slc_16afrisr0107_Lhh_tcal_test.rat": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lhh_tcal_test.rat",
    "data/polsar/slc_16afrisr0107_Lhh_tcal_test.rat.hdr": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lhh_tcal_test.rat.hdr",
    "data/polsar/slc_16afrisr0107_Lhv_tcal_test.rat": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lhv_tcal_test.rat",
    "data/polsar/slc_16afrisr0107_Lhv_tcal_test.rat.hdr": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lhv_tcal_test.rat.hdr",
    "data/polsar/slc_16afrisr0107_Lvh_tcal_test.rat": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lvh_tcal_test.rat",
    "data/polsar/slc_16afrisr0107_Lvh_tcal_test.rat.hdr": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lvh_tcal_test.rat.hdr",
    "data/polsar/slc_16afrisr0107_Lvv_tcal_test.rat": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lvv_tcal_test.rat",
    "data/polsar/slc_16afrisr0107_Lvv_tcal_test.rat.hdr": "https://zenodo.org/records/10797182/files/slc_16afrisr0107_Lvv_tcal_test.rat.hdr",
    "data/insar/common_profile.npy": "https://zenodo.org/records/10797182/files/common_profile.npy",
    "data/insar/flat_earth_Mondah_S_2015_11_11_cut.npy": "https://zenodo.org/records/10797182/files/flat_earth_Mondah_S_2015_11_11_cut.npy",
    "data/insar/flat_earth_Mondah_S_2015_11_11_cut.rat": "https://zenodo.org/records/10797182/files/flat_earth_Mondah_S_2015_11_11_cut.rat",
    "data/insar/image1.rat": "https://zenodo.org/records/10797182/files/image1.rat",
    "data/insar/image2.rat": "https://zenodo.org/records/10797182/files/image2.rat",
    "data/insar/master_HH_Mondah_S_2015_11_11_cut.npy": "https://zenodo.org/records/10797182/files/master_HH_Mondah_S_2015_11_11_cut.npy",
    "data/insar/master_HH_Mondah_S_2015_11_11_cut.rat": "https://zenodo.org/records/10797182/files/master_HH_Mondah_S_2015_11_11_cut.rat",
    "data/insar/RH100_Mondah_S_2015_11_11_cut.npy": "https://zenodo.org/records/10797182/files/RH100_Mondah_S_2015_11_11_cut.npy",
    "data/insar/RH100_Mondah_S_2015_11_11_cut.rat": "https://zenodo.org/records/10797182/files/RH100_Mondah_S_2015_11_11_cut.rat",
    "data/insar/slave_HH_Mondah_S_2015_11_11_cut.npy": "https://zenodo.org/records/10797182/files/slave_HH_Mondah_S_2015_11_11_cut.npy",
    "data/insar/slave_HH_Mondah_S_2015_11_11_cut.rat": "https://zenodo.org/records/10797182/files/slave_HH_Mondah_S_2015_11_11_cut.rat",
    "data/polinsar/gamma_max.rat": "https://zenodo.org/records/10797182/files/gamma_max.rat",
    "data/polinsar/gamma_min.rat": "https://zenodo.org/records/10797182/files/gamma_min.rat",
    "data/polinsar/incidence_15tmpsar0302_L_t01.rat": "https://zenodo.org/records/10797182/files/incidence_15tmpsar0302_L_t01.rat",
    "data/polinsar/kz_2d_demc_15tmpsar0302_15tmpsar0303_t01.rat": "https://zenodo.org/records/10797182/files/kz_2d_demc_15tmpsar0302_15tmpsar0303_t01.rat",
    "data/polinsar/Lida_r1503.rat": "https://zenodo.org/records/10797182/files/Lida_r1503.rat",
    "data/polinsar/local_slope_15tmpsar0302_L_t01.rat": "https://zenodo.org/records/10797182/files/local_slope_15tmpsar0302_L_t01.rat",
    "data/polinsar/pha_flat_15tmpsar0302_15tmpsar0303_Lhh_t01.rat": "https://zenodo.org/records/10797182/files/pha_flat_15tmpsar0302_15tmpsar0303_Lhh_t01.rat",
    "data/polinsar/pha_max.rat": "https://zenodo.org/records/10797182/files/pha_max.rat",
    "data/polinsar/pha_min.rat": "https://zenodo.org/records/10797182/files/pha_min.rat",
    "data/polinsar/slc_15tmpsar0302_Lhh_t01.rat": "https://zenodo.org/records/10797182/files/slc_15tmpsar0302_Lhh_t01.rat",
    "data/polinsar/slc_15tmpsar0302_Lhv_t01.rat": "https://zenodo.org/records/10797182/files/slc_15tmpsar0302_Lhv_t01.rat",
    "data/polinsar/slc_15tmpsar0302_Lvh_t01.rat": "https://zenodo.org/records/10797182/files/slc_15tmpsar0302_Lvh_t01.rat",
    "data/polinsar/slc_15tmpsar0302_Lvv_t01.rat": "https://zenodo.org/records/10797182/files/slc_15tmpsar0302_Lvv_t01.rat",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lhh_t01.rat": "https://zenodo.org/records/10797182/files/slc_coreg_15tmpsar0302_15tmpsar0303_Lhh_t01.rat",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lhv_t01.rat": "https://zenodo.org/records/10797182/files/slc_coreg_15tmpsar0302_15tmpsar0303_Lhv_t01.rat",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lvh_t01.rat": "https://zenodo.org/records/10797182/files/slc_coreg_15tmpsar0302_15tmpsar0303_Lvh_t01.rat",
    "data/polinsar/slc_coreg_15tmpsar0302_15tmpsar0303_Lvv_t01.rat": "https://zenodo.org/records/10797182/files/slc_coreg_15tmpsar0302_15tmpsar0303_Lvv_t01.rat",
    "data/tomosar/dem.rat": "https://zenodo.org/records/10797182/files/dem.rat",
    "data/tomosar/kz_00.rat": "https://zenodo.org/records/10797182/files/kz_00.rat",
    "data/tomosar/kz_01.rat": "https://zenodo.org/records/10797182/files/kz_01.rat",
    "data/tomosar/kz_02.rat": "https://zenodo.org/records/10797182/files/kz_02.rat",
    "data/tomosar/kz_03.rat": "https://zenodo.org/records/10797182/files/kz_03.rat",
    "data/tomosar/kz_04.rat": "https://zenodo.org/records/10797182/files/kz_04.rat",
    "data/tomosar/kz_05.rat": "https://zenodo.org/records/10797182/files/kz_05.rat",
    "data/tomosar/kz_06.rat": "https://zenodo.org/records/10797182/files/kz_06.rat",
    "data/tomosar/kz_07.rat": "https://zenodo.org/records/10797182/files/kz_07.rat",
    "data/tomosar/kz_08.rat": "https://zenodo.org/records/10797182/files/kz_08.rat",
    "data/tomosar/kz_09.rat": "https://zenodo.org/records/10797182/files/kz_09.rat",
    "data/tomosar/kz_10.rat": "https://zenodo.org/records/10797182/files/kz_10.rat",
    "data/tomosar/lidar_chm.rat": "https://zenodo.org/records/10797182/files/lidar_chm.rat",
    "data/tomosar/lidar_dtm.rat": "https://zenodo.org/records/10797182/files/lidar_dtm.rat",
    "data/tomosar/pha_dem_00.rat": "https://zenodo.org/records/10797182/files/pha_dem_00.rat",
    "data/tomosar/pha_dem_01.rat": "https://zenodo.org/records/10797182/files/pha_dem_01.rat",
    "data/tomosar/pha_dem_02.rat": "https://zenodo.org/records/10797182/files/pha_dem_02.rat",
    "data/tomosar/pha_dem_03.rat": "https://zenodo.org/records/10797182/files/pha_dem_03.rat",
    "data/tomosar/pha_dem_04.rat": "https://zenodo.org/records/10797182/files/pha_dem_04.rat",
    "data/tomosar/pha_dem_05.rat": "https://zenodo.org/records/10797182/files/pha_dem_05.rat",
    "data/tomosar/pha_dem_06.rat": "https://zenodo.org/records/10797182/files/pha_dem_06.rat",
    "data/tomosar/pha_dem_07.rat": "https://zenodo.org/records/10797182/files/pha_dem_07.rat",
    "data/tomosar/pha_dem_08.rat": "https://zenodo.org/records/10797182/files/pha_dem_08.rat",
    "data/tomosar/pha_dem_09.rat": "https://zenodo.org/records/10797182/files/pha_dem_09.rat",
    "data/tomosar/pha_dem_10.rat": "https://zenodo.org/records/10797182/files/pha_dem_10.rat",
    "data/tomosar/pos_scatters_ground.npy": "https://zenodo.org/records/10797182/files/pos_scatters_ground.npy",
    "data/tomosar/pos_scatters_layer1.npy": "https://zenodo.org/records/10797182/files/pos_scatters_layer1.npy",
    "data/tomosar/slc_00_HH.rat": "https://zenodo.org/records/10797182/files/slc_00_HH.rat",
    "data/tomosar/slc_00_HV.rat": "https://zenodo.org/records/10797182/files/slc_00_HV.rat",
    "data/tomosar/slc_00_VV.rat": "https://zenodo.org/records/10797182/files/slc_00_VV.rat",
    "data/tomosar/slc_01_HH.rat": "https://zenodo.org/records/10797182/files/slc_01_HH.rat",
    "data/tomosar/slc_01_HV.rat": "https://zenodo.org/records/10797182/files/slc_01_HV.rat",
    "data/tomosar/slc_01_VV.rat": "https://zenodo.org/records/10797182/files/slc_01_VV.rat",
    "data/tomosar/slc_02_HH.rat": "https://zenodo.org/records/10797182/files/slc_02_HH.rat",
    "data/tomosar/slc_02_HV.rat": "https://zenodo.org/records/10797182/files/slc_02_HV.rat",
    "data/tomosar/slc_02_VV.rat": "https://zenodo.org/records/10797182/files/slc_02_VV.rat",
    "data/tomosar/slc_03_HH.rat": "https://zenodo.org/records/10797182/files/slc_03_HH.rat",
    "data/tomosar/slc_03_HV.rat": "https://zenodo.org/records/10797182/files/slc_03_HV.rat",
    "data/tomosar/slc_03_VV.rat": "https://zenodo.org/records/10797182/files/slc_03_VV.rat",
    "data/tomosar/slc_04_HH.rat": "https://zenodo.org/records/10797182/files/slc_04_HH.rat",
    "data/tomosar/slc_04_HV.rat": "https://zenodo.org/records/10797182/files/slc_04_HV.rat",
    "data/tomosar/slc_04_VV.rat": "https://zenodo.org/records/10797182/files/slc_04_VV.rat",
    "data/tomosar/slc_05_HH.rat": "https://zenodo.org/records/10797182/files/slc_05_HH.rat",
    "data/tomosar/slc_05_HV.rat": "https://zenodo.org/records/10797182/files/slc_05_HV.rat",
    "data/tomosar/slc_05_VV.rat": "https://zenodo.org/records/10797182/files/slc_05_VV.rat",
    "data/tomosar/slc_06_HH.rat": "https://zenodo.org/records/10797182/files/slc_06_HH.rat",
    "data/tomosar/slc_06_HV.rat": "https://zenodo.org/records/10797182/files/slc_06_HV.rat",
    "data/tomosar/slc_06_VV.rat": "https://zenodo.org/records/10797182/files/slc_06_VV.rat",
    "data/tomosar/slc_07_HH.rat": "https://zenodo.org/records/10797182/files/slc_07_HH.rat",
    "data/tomosar/slc_07_HV.rat": "https://zenodo.org/records/10797182/files/slc_07_HV.rat",
    "data/tomosar/slc_07_VV.rat": "https://zenodo.org/records/10797182/files/slc_07_VV.rat",
    "data/tomosar/slc_08_HH.rat": "https://zenodo.org/records/10797182/files/slc_08_HH.rat",
    "data/tomosar/slc_08_HV.rat": "https://zenodo.org/records/10797182/files/slc_08_HV.rat",
    "data/tomosar/slc_08_VV.rat": "https://zenodo.org/records/10797182/files/slc_08_VV.rat",
    "data/tomosar/slc_09_HH.rat": "https://zenodo.org/records/10797182/files/slc_09_HH.rat",
    "data/tomosar/slc_09_HV.rat": "https://zenodo.org/records/10797182/files/slc_09_HV.rat",
    "data/tomosar/slc_09_VV.rat": "https://zenodo.org/records/10797182/files/slc_09_VV.rat",
    "data/tomosar/slc_10_HH.rat": "https://zenodo.org/records/10797182/files/slc_10_HH.rat",
    "data/tomosar/slc_10_HV.rat": "https://zenodo.org/records/10797182/files/slc_10_HV.rat",
    "data/tomosar/slc_10_VV.rat": "https://zenodo.org/records/10797182/files/slc_10_VV.rat",
    "data/DInSAR.zip": "https://zenodo.org/records/11171238/files/DInSAR.zip",

}


legacy_registry = {
    ('data/' + filename): registry['data/' + filename] for filename in legacy_datasets
}



# from tempfile import NamedTemporaryFile

# import numpy as np
# import pytest

# from pysarpro.io import imread, imsave

# def test_prefered_plugin():
#     # Don't call use_plugin("imageio") before, this way we test that imageio is used
#     # by default
#     order = plugin_order()
#     assert order["imread"][0] == "imageio"
#     assert order["imsave"][0] == "imageio"
#     assert order["imread_collection"][0] == "imageio"


# class TestSave:
#     @pytest.mark.parametrize(
#         "shape,dtype",
#         [
#             # float32, float64 can't be saved as PNG and raise
#             # uint32 is not roundtripping properly
#             ((10, 10), np.uint8),
#             ((10, 10), np.uint16),
#             ((10, 10, 2), np.uint8),
#             ((10, 10, 3), np.uint8),
#             ((10, 10, 4), np.uint8),
#         ],
#     )
#     def test_imsave_roundtrip(self, shape, dtype, tmp_path):
#         if np.issubdtype(dtype, np.floating):
#             min_ = 0
#             max_ = 1
#         else:
#             min_ = 0
#             max_ = np.iinfo(dtype).max
#         expected = np.linspace(
#             min_, max_, endpoint=True, num=np.prod(shape), dtype=dtype
#         )
#         expected = expected.reshape(shape)
#         file_path = tmp_path / "roundtrip.png"
#         imsave(file_path, expected)
#         actual = imread(file_path)
#         np.testing.assert_array_almost_equal(actual, expected)

#     def test_bool_array_save(self):
#         with NamedTemporaryFile(suffix='.png') as f:
#             fname = f.name

#         with pytest.warns(UserWarning, match=r'.* is a boolean image'):
#             a = np.zeros((5, 5), bool)
#             a[2, 2] = True
#             imsave(fname, a)



import os
import sys

import click
from spin import util
from spin.cmds import meson


@click.command()
@click.argument("sphinx_target", default="html")
@click.option(
    "--clean",
    is_flag=True,
    default=False,
    help="Clean previously built docs before building",
)
@click.option(
    "--build/--no-build",
    "first_build",
    default=True,
    help="Build project before generating docs",
)
@click.option(
    "--plot/--no-plot",
    "sphinx_gallery_plot",
    default=True,
    help="Sphinx gallery: enable/disable plots",
)
@click.option("--jobs", "-j", default="auto", help="Number of parallel build jobs")
@click.option(
    "--install-deps/--no-install-deps",
    default=False,
    help="Install dependencies before building",
)
@click.pass_context
def docs(
    ctx, sphinx_target, clean, first_build, jobs, sphinx_gallery_plot, install_deps
):
    """📖 Build documentation

    By default, SPHINXOPTS="-W", raising errors on warnings.
    To build without raising on warnings:

      SPHINXOPTS="" spin docs

    The command is roughly equivalent to `cd doc && make SPHINX_TARGET`.
    To get a list of viable `SPHINX_TARGET`:

      spin docs help

    """
    if install_deps:
        util.run(['pip', 'install', '-q', '-r', 'requirements/docs.txt'])

    for extra_param in ('install_deps',):
        del ctx.params[extra_param]
    ctx.forward(meson.docs)


@click.command()
@click.argument("asv_args", nargs=-1)
def asv(asv_args):
    """🏃 Run `asv` to collect benchmarks

    ASV_ARGS are passed through directly to asv, e.g.:

    spin asv -- dev -b TransformSuite

    Please see CONTRIBUTING.txt
    """
    site_path = meson._get_site_packages()
    if site_path is None:
        print("No built scikit-image found; run `spin build` first.")
        sys.exit(1)

    os.environ['PYTHONPATH'] = f'{site_path}{os.sep}:{os.environ.get("PYTHONPATH", "")}'
    util.run(['asv'] + list(asv_args))


@click.command()
def sdist():
    """📦 Build a source distribution in `dist/`."""
    util.run(['python', '-m', 'build', '.', '--sdist'])


@click.command(context_settings={'ignore_unknown_options': True})
@click.argument("ipython_args", metavar='', nargs=-1)
@click.pass_context
def ipython(ctx, ipython_args):
    """💻 Launch IPython shell with PYTHONPATH set

    OPTIONS are passed through directly to IPython, e.g.:

    spin ipython -i myscript.py
    """
    env = os.environ
    env['PYTHONWARNINGS'] = env.get('PYTHONWARNINGS', 'all')

    preimport = (
        r"import pysarpro as sarpro; "
        r"print(f'\nPreimported pysarpro {sarpro.__version__} as sarpro')"
    )
    ctx.params['ipython_args'] = (
        f"--TerminalIPythonApp.exec_lines={preimport}",
    ) + ipython_args

    ctx.forward(meson.ipython)



"""Standard test images.

For more images, see

 - http://sipi.usc.edu/database/database.php

"""

import os
import os.path as osp
import re
import shutil
import zipfile

from .. import __version__
from ._registry import registry, registry_urls

_LEGACY_DATA_DIR = osp.dirname(__file__)
_DISTRIBUTION_DIR = osp.dirname(_LEGACY_DATA_DIR)

try:
    from pooch import file_hash
except ModuleNotFoundError:
    # Function taken from
    # https://github.com/fatiando/pooch/blob/master/pooch/utils.py
    def file_hash(fname, alg="sha256"):
        """
        Calculate the hash of a given file.
        Useful for checking if a file has changed or been corrupted.
        Parameters
        ----------
        fname : str
            The name of the file.
        alg : str
            The type of the hashing algorithm
        Returns
        -------
        hash : str
            The hash of the file.
        Examples
        --------
        >>> fname = "test-file-for-hash.txt"
        >>> with open(fname, "w") as f:
        ...     __ = f.write("content of the file")
        >>> print(file_hash(fname))
        0fc74468e6a9a829f103d069aeb2bb4f8646bad58bf146bb0e3379b759ec4a00
        >>> import os
        >>> os.remove(fname)
        """
        import hashlib

        if alg not in hashlib.algorithms_available:
            raise ValueError(f'Algorithm \'{alg}\' not available in hashlib')
        # Calculate the hash in chunks to avoid overloading the memory
        chunksize = 65536
        hasher = hashlib.new(alg)
        with open(fname, "rb") as fin:
            buff = fin.read(chunksize)
            while buff:
                hasher.update(buff)
                buff = fin.read(chunksize)
        return hasher.hexdigest()


def _has_hash(path, expected_hash):
    """Check if the provided path has the expected hash."""
    if not osp.exists(path):
        return False
    return file_hash(path) == expected_hash


def _create_image_fetcher():
    try:
        import pooch

        # older versions of Pooch don't have a __version__ attribute
        if not hasattr(pooch, '__version__'):
            retry = {}
        else:
            retry = {'retry_if_failed': 3}
    except ImportError:
        # Without pooch, fallback on the standard data directory
        # which for now, includes a few limited data samples
        return None, _LEGACY_DATA_DIR

    # Pooch expects a `+` to exist in development versions.
    # Since pysarpro doesn't follow that convention, we have to manually
    # remove `.dev` with a `+` if it exists.
    # This helps pooch understand that it should look in master
    # to find the required files
    if '+git' in __version__:
        pysarpro_version_for_pooch = __version__.replace('.dev0+git', '+git')
    else:
        pysarpro_version_for_pooch = __version__.replace('.dev', '+')

    if '+' in pysarpro_version_for_pooch:
        url = "https://github.com/Pol-InSAR/pysarpro/raw/" "{version}/pysarpro/"
    else:
        url = "https://github.com/Pol-InSAR/pysarpro/raw/" "v{version}/pysarpro/"

    # Create a new friend to manage your sample data storage
    image_fetcher = pooch.create(
        # Pooch uses appdirs to select an appropriate directory for the cache
        # on each platform.
        # https://github.com/ActiveState/appdirs
        # On linux this converges to
        # '$HOME/.cache/pysarpro'
        # With a version qualifier
        path=pooch.os_cache("pysarpro"),
        base_url=url,
        version=pysarpro_version_for_pooch,
        version_dev="main",
        env="pysarpro_DATADIR",
        registry=registry,
        urls=registry_urls,
        # Note: this should read `retry_if_failed=3,`, but we generate that
        # dynamically at import time above, in case installed pooch is a less
        # recent version
        **retry,
    )

    data_dir = osp.join(str(image_fetcher.abspath), 'data')
    return image_fetcher, data_dir


_image_fetcher, data_dir = _create_image_fetcher()


def _skip_pytest_case_requiring_pooch(data_filename):
    """If a test case is calling pooch, skip it.

    This running the test suite in environments without internet
    access, skipping only the tests that try to fetch external data.
    """

    # Check if pytest is currently running.
    # Packagers might use pytest to run the tests suite, but may not
    # want to run it online with pooch as a dependency.
    # As such, we will avoid failing the test, and silently skipping it.
    if 'PYTEST_CURRENT_TEST' in os.environ:
        # https://docs.pytest.org/en/latest/example/simple.html#pytest-current-test-environment-variable  #noqa
        import pytest

        # Pytest skip raises an exception that allows the
        # tests to be skipped
        pytest.skip(f'Unable to download {data_filename}', allow_module_level=True)


def _ensure_cache_dir(*, target_dir):
    """Prepare local cache directory if it doesn't exist already.

    Creates::

        /path/to/target_dir/
                 └─ data/
                    └─ README.txt
    """
    os.makedirs(osp.join(target_dir, "data"), exist_ok=True)
    readme_src = osp.join(_DISTRIBUTION_DIR, "data/README.txt")
    readme_dest = osp.join(target_dir, "data/README.txt")
    if not osp.exists(readme_dest):
        shutil.copy2(readme_src, readme_dest)


def _fetch(data_filename):
    """Fetch a given data file from either the local cache or the repository.

    This function provides the path location of the data file given
    its name in the pysarpro repository. If a data file is not included in the
    distribution and pooch is available, it is downloaded and cached.

    Parameters
    ----------
    data_filename : str
        Name of the file in the pysarpro repository. e.g.
        'restoration/tess/camera_rl.npy'.

    Returns
    -------
    file_path : str
        Path of the local file.

    Raises
    ------
    KeyError:
        If the filename is not known to the pysarpro distribution.

    ModuleNotFoundError:
        If the filename is known to the pysarpro distribution but pooch
        is not installed.

    ConnectionError:
        If pysarpro is unable to connect to the internet but the
        dataset has not been downloaded yet.
    """
    expected_hash = registry[data_filename]
    if _image_fetcher is None:
        cache_dir = osp.dirname(data_dir)
    else:
        cache_dir = str(_image_fetcher.abspath)

    # Case 1: the file is already cached in `data_cache_dir`
    cached_file_path = osp.join(cache_dir, data_filename)
    if _has_hash(cached_file_path, expected_hash):
        # Nothing to be done, file is where it is expected to be
        return cached_file_path

    # Case 2: file is present in `legacy_data_dir`
    legacy_file_path = osp.join(_DISTRIBUTION_DIR, data_filename)
    if _has_hash(legacy_file_path, expected_hash):
        return legacy_file_path

    # Case 3: file is not present locally
    if _image_fetcher is None:
        _skip_pytest_case_requiring_pooch(data_filename)
        raise ModuleNotFoundError(
            "The requested file is part of the pysarpro distribution, "
            "but requires the installation of an optional dependency, pooch. "
            "To install pooch, use your preferred python package manager. "
            "Follow installation instruction found at "
            "https://pol-insar.github.io/docs/stable/user_guide/install.html"
        )
    # Download the data with pooch which caches it automatically
    _ensure_cache_dir(target_dir=cache_dir)
    try:
        cached_file_path = _image_fetcher.fetch(data_filename)
        return cached_file_path
    except ConnectionError as err:
        _skip_pytest_case_requiring_pooch(data_filename)
        # If we decide in the future to suppress the underlying 'requests'
        # error, change this to `raise ... from None`. See PEP 3134.
        raise ConnectionError(
            'Tried to download a pysarpro dataset, but no internet '
            'connection is available. To avoid this message in the '
            'future, try `pysarpro.data.download_all()` when you are '
            'connected to the internet.'
        ) from err


def download_all(directory=None, pattern=None):
    """Download all datasets for use with pysarpro offline.

    pysarpro datasets are no longer shipped with the library by default.
    This allows us to use higher quality datasets, while keeping the
    library download size small.

    This function requires the installation of an optional dependency, pooch,
    to download the full dataset. Follow installation instruction found at

        https://pol-insar.github.io/docs/stable/user_guide/install.html

    Call this function to download all sample images making them available
    offline on your machine.

    Parameters
    ----------
    directory: path-like, optional
        The directory where the dataset should be stored.
    pattern: str, optional
        A regex pattern to filter the datasets to be downloaded.

    Raises
    ------
    ModuleNotFoundError:
        If pooch is not install, this error will be raised.

    Notes
    -----
    pysarpro will only search for images stored in the default directory.
    Only specify the directory if you wish to download the images to your own
    folder for a particular reason. You can access the location of the default
    data directory by inspecting the variable ``pysarpro.data.data_dir``.
    """

    if _image_fetcher is None:
        raise ModuleNotFoundError(
            "To download all package data, pysarpro needs an optional "
            "dependency, pooch."
            "To install pooch, follow our installation instructions found at "
            "https://pol-insar.github.io/docs/stable/user_guide/install.html"
        )
    # Consider moving this kind of logic to Pooch
    old_dir = _image_fetcher.path
    try:
        if directory is not None:
            directory = osp.expanduser(directory)
            _image_fetcher.path = directory
        _ensure_cache_dir(target_dir=_image_fetcher.path)

        for data_filename in _image_fetcher.registry:
            if pattern is not None and not re.search(pattern, data_filename):
                continue
            file_path = _fetch(data_filename)

            # Copy to `directory` or implicit cache if it is not already there
            if not file_path.startswith(str(_image_fetcher.path)):
                dest_path = osp.join(_image_fetcher.path, data_filename)
                os.makedirs(osp.dirname(dest_path), exist_ok=True)
                shutil.copy2(file_path, dest_path)

            # If the file is a zip file, unzip it and remove the zipped file
            if data_filename.endswith('.zip'):
                with zipfile.ZipFile(dest_path, 'r') as zip_ref:
                    zip_ref.extractall(osp.dirname(dest_path))
                os.remove(dest_path)
    finally:
        _image_fetcher.path = old_dir


def lbp_frontal_face_cascade_filename():
    """Return the path to the XML file containing the weak classifier cascade.

    These classifiers were trained using LBP features. The file is part
    of the OpenCV repository [1]_.

    References
    ----------
    .. [1] OpenCV lbpcascade trained files
           https://github.com/opencv/opencv/tree/master/data/lbpcascades
    """

    return _fetch('data/lbpcascade_frontalface_opencv.xml')


def _load(f, as_gray=False):
    """Load an image file located in the data directory.

    Parameters
    ----------
    f : string
        File name.
    as_gray : bool, optional
        Whether to convert the image to grayscale.

    Returns
    -------
    img : ndarray
        Image loaded from ``pysarpro.data_dir``.
    """
    # importing io is quite slow since it scans all the backends
    # we lazy import it here
    from ..io import imread

    return imread(_fetch(f), as_gray=as_gray)


def astronaut():
    """Color image of the astronaut Eileen Collins.

    Photograph of Eileen Collins, an American astronaut. She was selected
    as an astronaut in 1992 and first piloted the space shuttle STS-63 in
    1995. She retired in 2006 after spending a total of 38 days, 8 hours
    and 10 minutes in outer space.

    This image was downloaded from the NASA Great Images database
    <https://flic.kr/p/r9qvLn>`__.

    No known copyright restrictions, released into the public domain.

    Returns
    -------
    astronaut : (512, 512, 3) uint8 ndarray
        Astronaut image.
    """

    return _load("data/astronaut.png")
